def error_msg_default(prompt, result, error):
    """Default function to append the error message to the prompt.

    Args:
        prompt (list): List of messages in OpenAI format.
        result (str): Result of the LLM query.
        error (str): Error message for the LLM.
    
    Returns:
        prompt (list): List of messages in OpenAI format.
    """
    prompt.append({"role":"assistant", "content":result})
    prompt.append({"role":"user", "content":error})
    return prompt


try:
    import openai
except ImportError:
    raise ImportError("Please install openai to use built-in LLM API.")
from openai import OpenAI, AzureOpenAI
import time
import os
from .utils import match_model
from .base import BaseModel

def initialize_client():
    if os.environ.get("OPENAI_KEY"):
        return OpenAI(
            api_key=os.environ["OPENAI_KEY"],
            organization=os.environ["OPENAI_ORG"],
        )
    elif os.environ.get("AZURE_OPENAI_API_KEY"):
        return AzureOpenAI(
            api_key=os.environ["AZURE_OPENAI_API_KEY"],
            api_version=os.environ["AZURE_OPENAI_API_VERSION"],
            azure_endpoint=os.environ["AZURE_OPENAI_ENDPOINT"],
        ), os.environ.get("AZURE_DEPLOYMENT_NAME")

    openai_key_path = os.path.join(os.path.expanduser('~'), ".openai", "openai.key")
    azure_key_path = os.path.join(os.path.expanduser('~'), ".openai", "azure_openai.key")

    if os.path.exists(openai_key_path):
        with open(openai_key_path, 'r') as f:
            lines = f.read().splitlines()
            if len(lines) >= 2:
                return OpenAI(
                    api_key=lines[0].strip(),
                    organization=lines[1].strip(),
                )
        raise FileNotFoundError(
            """
            Please create a file at ~/.openai/openai.key with your OpenAI API key and organization ID. 
            The first line should be your API key and the second line should be your organization ID.
            """
        )
    
    if os.path.exists(azure_key_path):
        with open(azure_key_path, 'r') as f:
            lines = f.read().splitlines()
            if len(lines) >= 3:
                return AzureOpenAI(
                    api_key=lines[0].strip(),
                    api_version=lines[1].strip(),
                    azure_endpoint=lines[2].strip(),
                ), lines[3].strip()

        raise FileNotFoundError(
            """
            Please create a file at ~/.openai/azure_openai.key with your Azure OpenAI API key, 
            API version, Azure endpoint, and deployment name. 
            The first line should be your API key, the second line should be your API version, 
            the third line should be your Azure endpoint, and the fourth line should be your deployment name.
            """
        )

client_and_deployment = initialize_client()
client = client_and_deployment[0] if isinstance(client_and_deployment, tuple) else client_and_deployment
deployment_name = client_and_deployment[1] if isinstance(client_and_deployment, tuple) else None

class GPT_chat(BaseModel):

    def __init__(self, model_name, global_counter=None):
        model_max, enc_fn = match_model(model_name)
        self.enc_fn = enc_fn
        self.model_max = model_max
        super().__init__(model_name, global_counter, "chat")
    
    def encode(self, txt):
        return self.enc_fn.encode(txt)

    def decode(self, txt):
        return self.enc_fn.decode(txt)

    def query_chat(self, messages, shrink_idx, max_gen=512, temp=0.):
        model_max = self.model_max
        messages = self.shrink_msg(messages, shrink_idx, model_max-max_gen)
        while(True):
            try:
                if isinstance(client, OpenAI):
                    completion = client.chat.completions.create(
                        model=self.name,
                        messages=messages,
                        temperature=temp,
                        max_tokens=max_gen,
                    )
                elif isinstance(client, AzureOpenAI):
                    completion = client.completions.create(
                        model=deployment_name,
                        messages=messages,
                        temperature=temp,
                        max_tokens=max_gen,
                    )
                return completion.choices[0].message.content, {"prompt":completion.usage.prompt_tokens, "completion":completion.usage.completion_tokens, "total":completion.usage.total_tokens}
            except Exception as e:
                if self.debug:
                    raise e
                elif isinstance(e, openai.RateLimitError) or isinstance(e, openai.APIStatusError) or isinstance(e, openai.APITimeoutError) or isinstance(e, openai.APIConnectionError) or isinstance(e, openai.InternalServerError):
                    time.sleep(30)
                    print(e)
                elif "However, your messages resulted in" in str(e):
                    print("error:", e, str(e))
                    e = str(e)
                    index = e.find("your messages resulted in ")
                    import re
                    val = int(re.findall(r'\d+', e[index + len("your messages resulted in ") : ])[0])
                    index2 = e.find("maximum context length is ")
                    model_max = int(re.findall(r'\d+', e[index2 + len("maximum context length is "):])[0])
                    messages = self.shrink_msg_by(messages, shrink_idx, val-model_max)
                else:
                    print("error:", e)
                    print("retrying in 5 seconds")
                    time.sleep(5)



from .exceptions import AfterQueryError
from collections.abc import Callable, Awaitable
from .node_functions import error_msg_default
from .base_node import BaseNode
from .graph import Graph
from colorama import Fore, Back, Style
import typing as t

from .after_query import BaseAfterQuery
from .compose_prompt import ComposePromptDB

class SimpleDBNode(BaseNode):
    """Class for a node in the graph that queries a database.

    Each node in the graph is an instance of the SimpleDBNode class. The node is evaluated by querying the LLM with a prompt.

    Attributes:
        db (Any): Database object.
    """
    def __init__(self, key:str, prompt: str, graph: Graph, query_llm: Callable, compose_prompt: ComposePromptDB, database: t.Any, after_query: BaseAfterQuery = None, error_msg_fn: Callable[[list, str, AfterQueryError], list] = error_msg_default, verbose: bool = False, token_counter: Callable = None):
        """Initializes the SimpleDBNode class.

        Args:
            key (str): Unique key for the node.
            prompt (str): Prompt for the node.
            graph (Graph): Graph object.
            query_llm (Callable): Function to query the LLM.
            compose_prompt (ComposePromptDB): ComposePromptDB object.
            database (Any): Database object.
            after_query (BaseAfterQuery): AfterQuery object.
            error_msg_fn (Callable): Function to add error message to the prompt.
            verbose (bool): Verbose flag.
            token_counter (Callable): Function to count tokens.
        """
        super().__init__(key, prompt, graph, query_llm, compose_prompt, after_query=after_query, error_msg_fn=error_msg_fn, verbose=verbose, token_counter=token_counter)
        self.db: t.Any = database
        self._compose_prompt.set_node(self)
        self.rendered_prompt = None
    
    def compose_prompt(self):
        return self._compose_prompt(dependencies=self.get_dependencies(), prompt=self.prompt)

    def _print_answer(self, msg):
        if self.verbose:
            if hasattr(self, 'db_retrieval_results') and len(self.db_retrieval_results) > 0:
                print("DB operations: " + Style.DIM + Fore.BLUE + "{}".format(self.db_retrieval_results) + Style.RESET_ALL)
            print("Answer: " + Style.DIM + "{}".format(msg) + Style.RESET_ALL)


from .exceptions import AfterQueryError
from .node_functions import error_msg_default
from collections.abc import Callable, Awaitable
from .graph import Graph
from .after_query import BaseAfterQuery
from .compose_prompt import BaseComposePrompt
from colorama import Fore, Back, Style
import copy
import datetime
try:
    from wandb.sdk.data_types.trace_tree import Trace
except:
    pass

class BaseNode:
    """Base class for a node in the graph.

    Each node in the graph is an instance of the BaseNode class. The node is evaluated by querying the LLM with a prompt.

    Attributes:
        key (str): Unique key for the node.
        prompt (str): Prompt for the node.
        result (str): Result of the node evaluation.
        temporary_skip (bool): Flag to skip the node evaluation.
        graph (Graph): Graph object.
        adjacent_to (list): List of nodes that are adjacent to this node.
        adjacent_from (list): List of nodes that are adjacent from this node.
        evaluate_after (list): List of nodes that are evaluated after this node.
        counts (list): List of token counts.
        query_llm (Callable): Function to query the LLM.
        _compose_prompt (BaseComposePrompt): ComposePrompt object.
        after_query (BaseAfterQuery): AfterQuery object.
        error_msg_fn (Callable): Function to add error message to the prompt.
        verbose (bool): Verbose flag.
        token_counter (Callable): Function to count tokens.
    """
    def __init__(self, key:str, prompt:str, graph:Graph, query_llm:Callable, compose_prompt:BaseComposePrompt, after_query:BaseAfterQuery=None, error_msg_fn:Callable[[list,str,AfterQueryError],list]=error_msg_default, verbose:bool=False, token_counter:Callable=None):
        """Initializes the BaseNode class.
        
        Args:
            key (str): Unique key for the node.
            prompt (str): Prompt for the node.
            graph (Graph): Graph object.
            query_llm (Callable): Function to query the LLM.
            compose_prompt (BaseComposePrompt): ComposePrompt object.
            after_query (BaseAfterQuery): AfterQuery object.
            error_msg_fn (Callable): Function to add error message to the prompt.
            verbose (bool): Verbose flag.
            token_counter (Callable): Function to count tokens.
        """
        self.key = key
        self.prompt = prompt
        self.result = None
        self.temporary_skip = False
        self.temporary_skip = False
        self.graph = graph
        self.adjacent_to = []  # this -> node
        self.adjacent_from = []  # node -> this
        self.evaluate_after = []  # node -> this
        self.counts = []
        self.query_llm = query_llm
        self._compose_prompt = compose_prompt
        self.after_query = None
        self.chain_span = None
        if after_query is not None:
            self.after_query = after_query
            self.after_query.set_node(self)
        self.verbose = verbose
        self._add_error_msg = error_msg_fn
        
        if token_counter is not None:
            self.token_counter = token_counter
        else:
            self.token_counter = None
    
    def get_dependencies(self):
        """Get the dependencies (edges) of the node.

        Returns:
            list: List of dependencies.
        """
        return self.adjacent_from
    
    def get_dependencies_inc_order(self):
        """Get the dependencies (edges) and order (non-dependency edges) of the node.
        
        Returns:
            list: List of dependencies.
        """
        return self.adjacent_from + self.evaluate_after

    def compose_prompt(self):
        return self._compose_prompt(dependencies=self.get_dependencies(), prompt=self.prompt)

    def _print_question(self):
        if self.verbose:
            print(Fore.CYAN + "Prompt: " + Style.DIM + "{}".format(self.prompt) + Style.RESET_ALL)
    
    def _print_answer(self, msg):
        if self.verbose:
            print("Answer: " + Style.DIM + "{}".format(msg) + Style.RESET_ALL)
    
    def _print_skip(self):
        if self.verbose:
            print("Skip: " + Style.DIM + "{}".format(self.prompt) + Style.RESET_ALL)
    
    def skip_turn(self):
        """Temporarily skip the node evaluation. The previous result of the node will be reused.

        Raises:
            AssertionError: If the node has never been evaluated.
        """
        assert self.result is not None, "Attempting to skip a node ({}) that has never been evaluated".format(self.key)
        self.temporary_skip = True
    
    def _query_llm(self, prompt, shrink_idx):

        start_time_ms = round(datetime.datetime.now().timestamp() * 1000)
        result = self.query_llm(prompt, shrink_idx)
        end_time_ms = round(datetime.datetime.now().timestamp() * 1000)

        if self.chain_span is not None:
            llm_span = Trace(
                "OpenAI",
                kind="llm",
                start_time_ms=start_time_ms,
                end_time_ms=end_time_ms,
                inputs={"prompt":prompt},
                status_code="success",
                outputs={"response": result},
            )
            self.chain_span.add_child(llm_span)
        return result
    
    def _after_query(self, ignore_errors=False):
        if self.after_query is None:
            return
        status_code = "success"
        status_message = ""
        error = None
        after_query_input = self.result
        try:
            start_time_ms = round(datetime.datetime.now().timestamp() * 1000)
            self.after_query()
        except AfterQueryError as e:
            status_code = "error"
            status_message = e.error
            error = e
            self.result = after_query_input
        end_time_ms = round(datetime.datetime.now().timestamp() * 1000)
        if self.chain_span is not None:
            tool_span = Trace(
                "AfterQuery",
                kind="tool",
                status_code=status_code,
                status_message=status_message,
                start_time_ms=start_time_ms,
                end_time_ms=end_time_ms,
                inputs={"input": after_query_input},
                outputs={"response": self.result},
            )
            self.chain_span.add_child(tool_span)
        if status_code == "error":
            if ignore_errors:
                print(Fore.RED + "WARNING: ({}, {}) for {}".format(error, status_message, self.result) + Fore.RESET)
                self.result = "N/A"
            else:
                raise error

    def set_trace(self):
        if self.graph.chain_span is None:
            self.chain_span = None
            return
        chain_span = Trace(
            "NodeChain",
            kind="chain",
            start_time_ms=round(datetime.datetime.now().timestamp() * 1000),
            metadata={"prompt": self.prompt, "key": self.key},
        )
        self.chain_span = chain_span

    def commit_trace(self):
        if self.chain_span is None:
            return
        self.chain_span._span.end_time_ms = round(datetime.datetime.now().timestamp() * 1000)
        self.chain_span.add_inputs_and_outputs(inputs={"dependencies": [(node.key,node.result) for node in self.adjacent_from]}, 
                                               outputs={"response": self.result})
        self.graph.chain_span.add_child(self.chain_span)
        self.chain_span = None

    def evaluate(self):
        """Evaluate the node by querying the LLM.

        Retries the AfterQuery with the LLM up to 3 times in case of an error.

        Returns:
            str: Result of the node evaluation.

        Raises:
            AssertionError: If any dependency of the node has not been evaluated.
        """
        for node in self.adjacent_from:
            assert node.result is not None, "Dependency {} of {} has been not evaluated".format(node.key, self.key)

        if not self.temporary_skip:
            self.set_trace()
            prompt, shrink_idx = self.compose_prompt()
            error = None
            self._print_question()
            for i in range(3):
                try:
                    temp_prompt = copy.deepcopy(prompt)
                    if error is not None:
                        temp_prompt = self._add_error_msg(temp_prompt, self.result, error)
                    self.result, usage = self._query_llm(temp_prompt, shrink_idx)
                    if usage is not None:
                        self.counts.append(copy.copy(usage))
                    elif self.token_counter is not None:
                        self.counts.append({'prompt': self.token_counter(temp_prompt), 'completion':self.token_counter(self.result)})
                    self._after_query(ignore_errors=(i==2))
                    break
                except AfterQueryError as e:
                    error = e.error
            self.commit_trace()
            self._print_answer(self.result)
            print()
        else:
            self.temporary_skip = False
        return self.result

    def get_token_counts(self):
        """Get the LLM token counts for the specific node since instantiation.

        Returns:
            dict: Token counts for the node.
        """

        if self.token_counter is None:
            print("Warning: token counter not set for node {}".format(self.key))
        return {
            'calls': len(self.counts),
            'prompt': sum([c['prompt'] for c in self.counts]),
            'completion': sum([c['completion'] for c in self.counts])
        }


import numpy as np
try:
    import tiktoken
except ImportError:
    raise ImportError("Please install tiktoken to use built-in LLM API.")
import difflib


enc_fns = {
    "gpt-4":tiktoken.encoding_for_model("gpt-4"),
    "gpt-4-turbo":tiktoken.encoding_for_model("gpt-4"),
    "gpt-4-1106-preview":tiktoken.encoding_for_model("gpt-4"),
    "gpt-4-32k-0613":tiktoken.encoding_for_model("gpt-4"),
    "gpt-3.5-turbo":tiktoken.encoding_for_model("gpt-3.5-turbo"),
    "gpt-3.5-turbo-1106":tiktoken.encoding_for_model("gpt-3.5-turbo"),
    "gpt-3.5-turbo-0125":tiktoken.encoding_for_model("gpt-3.5-turbo"),
    "claude-3": None,
    "claude-2.1": None,
}
model_maxes = {
    "gpt-4":8192,
    "gpt-4-turbo":128000,
    "gpt-4-1106-preview":128000,
    "gpt-4-32k-0613":32768,
    "gpt-3.5-turbo":16385,
    "gpt-3.5-turbo-1106":16385,
    "gpt-3.5-turbo-0125":16385,
    "claude-3":50000,
    "claude-2.1":50000,
}

def match_model(model_name):

    # match model name to the closest string in enc_fns.keys()
    model_name = model_name.lower()
    matches = difflib.get_close_matches(model_name, enc_fns.keys())
    print("Matched model name to: ", matches)
    matches += [model for model in model_maxes.keys() if model_name.startswith(model)]
    if len(matches) == 0:
        raise ValueError("Model name {} not found!".format(model_name))
    else:
        model = matches[0]
    
    model_max = model_maxes[model]
    enc_fn = enc_fns[model]
    return model_max, enc_fn




class AfterQueryError(Exception):
    """Exception raised for errors in the after query postprocessing.
    
    Attributes:
        error (str): Error message.
    """
    def __init__(self, message, error):
        """Initializes the AfterQueryError class.

        Args:
            message (str): System error message.
            error (str): Error message for the LLM (a more descriptive error message for LLM to understand)
        """
        # Call the base class constructor with the parameters it needs
        super().__init__(message)
            
        # Set the error message
        self.error = error


import setuptools
import pkg_resources
import pathlib

PKG_NAME = "agentkit-llm"
VERSION = "0.1.6"
EXTRAS = {
    "logging": ["wandb"],
    "all": ["wandb", "openai", "anthropic", "tiktoken"],
}

setuptools.setup(
    name=PKG_NAME,
    version=VERSION,
    author="AgentKit Team",
    description='A LLM prompting framework for LLM agents',
    url="https://github.com/Holmeswww/AgentKit",
    long_description=pathlib.Path('README.md').read_text(),
    long_description_content_type='text/markdown',
    packages=setuptools.find_namespace_packages('src'),
    package_dir={'': 'src'},
    entry_points={'console_scripts': ['agentkit=agentkit.run_gui:main']},
    install_requires = ['colorama', 'numpy'],
    extras_require=EXTRAS,
    include_package_data=True,
    license="CC-BY-4.0-Attribution",
    classifiers=[
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
)



import re
class BaseComposePrompt:
    """Base class for composing prompts. This class is used to compose prompts for the LLM.

    Each compose prompt instance takes dependencies (QA results from dependent nodes) and 
    a prompt as input and returns a prompt in OpenAI format.

    Attributes:
        system_prompt (str): System prompt to be used in the conversation.
    """

    def __init__(self, system_prompt:str = "You are a helpful assistant."):
        """Initializes the BaseComposePrompt class.

        Args:
            system_prompt (str): System prompt to be used in the conversation.
        """
        self.system_prompt = system_prompt
    
    def compose(self, dependencies, prompt):
        """Compose prompt for the LLM.
        
        Args:
            dependencies (list): List of dependencies.
            prompt (str): User prompt.
        
        Returns:
            msg (list): List of messages in OpenAI format.
            shrink_idx (int): Index to shrink the prompt in case of truncation.
        """
        msg = [{"role": "system", "content": self.system_prompt}]
        msg = self.add_dependencies(msg, dependencies)
        msg.append({"role": "user", "content": prompt})
        return msg, 1

    def __call__(self, dependencies, prompt):
        return self.compose(dependencies, prompt)

    def add_dependencies(self, messages, dependencies):
        """Add dependencies to the messages.

        Args:
            messages (list): List of messages.
            dependencies (list): List of dependencies.
        
        Returns:
            messages (list): List of messages with dependencies in OpenAI format.
        """
        if len(dependencies)>0:
            for node in dependencies:
                messages.append({"role": "user", "content": node.prompt})
                messages.append({"role": "assistant", "content": node.result})
                assert type(messages[-1]['content']) == str, "Invalid type: {}".format(type(messages[-1]['content']))
        return messages

class ComposePromptDB(BaseComposePrompt):
    """Class for composing prompts with database values. This class is used to compose prompts for the LLM.

    Each compose prompt instance takes dependencies (QA results from dependent nodes) and
    a prompt as input and returns a prompt in OpenAI format.

    Attributes:
        system_prompt (str): System prompt to be used in the conversation.
        node (Node): Corresponding Node object.
    """
    
    def __init__(self, system_prompt:str = "You are a helpful assistant."):
        """Initializes the ComposePromptDB class.

        Args:
            system_prompt (str): System prompt to be used in the conversation.
        """
        super().__init__(system_prompt)
        self.node = None
    
    def set_node(self, node):
        """Set the node for the ComposePromptDB class.

        Args:
            node (Node): Corresponding Node object.
        """
        self.node = node
    
    def compose(self, dependencies, prompt):
        """Compose prompt for the LLM with database augmentation.

        Prompt text may contain placeholders for db values. This function will replace the placeholders with the actual values from db.

        Args:
            dependencies (list): List of dependencies.
            prompt (str): User prompt.
        
        Returns:
            msg (list): List of messages in OpenAI format.
            shrink_idx (int): Index to shrink the prompt in case of truncation.
        """

        msg = [{"role": "system", "content": self.system_prompt}]

        msg = self.add_dependencies(msg, dependencies, self.node.db)

        prompt, db_retrieval_results = self.render_db(prompt, self.node.db)
        self.node.rendered_prompt = prompt
        self.node.db_retrieval_results = db_retrieval_results

        msg.append({"role": "user", "content": prompt})

        return msg, self.shrink_idx


    def __call__(self, dependencies, prompt):

        if self.node is None:
            raise Exception("Node is not set")
        
        return self.compose(dependencies, prompt)
    
    def render_db(self, text, db):
        """Render db placeholders in the prompt text.

        Prompt text may contain placeholders for db values. This function will replace the placeholders with the actual values from db.

        Example:

        1. Confirm if the subgoal '$db.subgoals.subgoal$' is still accuracte for the challenge.

        2. Confirm if the subgoal '$db.subgoals.subgoal$' is incomplete and up-to-date according to the completion criteria '$db.subgoals.guide$'.
        
        Args:
            text (str): Prompt text.
            db (dict): Database values.
        
        Returns:
            text (str): Prompt text with db placeholders replaced with actual values.
            db_retrieval_results (list): List of db retrieval results.
        """
        db_retrieval_results = []
        pattern = r'\$db\.(.*?)\$'
        matches = re.findall(pattern, text)
        for match in matches:
            value = db
            for key in match.split('.'):
                if key in value:
                    value = value[key]
                else:
                    value = None
                    break
            if value != "None":
                text = text.replace(f'$db.{match}$', str(value))
            else:
                text = text.replace(f'$db.{match}$', "'NA'")
            db_retrieval_results.append((match, value))

        return text, db_retrieval_results

    def add_dependencies(self, messages, dependencies, db):
        """Add dependencies to the messages with db values.

        Args:
            messages (list): List of messages.
            dependencies (list): List of dependencies.
            db (dict): Database values.
        
        Returns:
            messages (list): List of messages with dependencies in OpenAI format.
        """
        if len(dependencies)>0:
            for node in dependencies:
                if node.key in db['shorthands'].keys():
                    messages.append({"role": "system", "content": "{}:\n\n{}".format(db['shorthands'][node.key], node.result)})
                else:
                    assert node.rendered_prompt != None, "Rendered prompt is not set for node {}".format(node.key)
                    messages.append({"role": "user", "content": node.rendered_prompt})
                    messages.append({"role": "assistant", "content": node.result})
                assert type(messages[-1]['content']) == str, "Invalid type: {}".format(type(messages[-1]['content']))
        return messages


from .utils import extract_json_objects
from .exceptions import AfterQueryError

class BaseAfterQuery:
    """Base class for after query postprocessing.

    Each after query instance performs postprocessing after the LLM query.

    Attributes:
        node (BaseNode): Node object.
    """

    def __init__(self):
        self.node = None
    
    def set_node(self, node):
        """Set the node for the after query.

        Args:
            node (BaseNode): Node object.
        """
        self.node = node
        return self
    
    def post_process(self):
        """Post process the result of the LLM query.
        
        This method can be overridden by the derived class to perform postprocessing.
        """
        pass

    def __call__(self):
        if self.node is None:
            raise Exception("Node is not set")

        self.post_process()

class JsonAfterQuery(BaseAfterQuery):
    """Class for after query postprocessing of Json objects.

    Attributes:
        type (type): Type of the Json object.
        required_keys (list): List of required keys in the Json object.
        length (int): Required length of the Json object.
    """

    def __init__(self):
        super().__init__()
        self.type = dict
        self.required_keys = []
        self.length = None

    def parse_json(self):
        """Parse the result of the LLM query.

        This method parses self.node.result (str) and returns the parsed Json object.
        The method checks the length of the parsed Json object, and if the Json object contains the required keys.

        Raises:
            AfterQueryError: If the answer is invalid.

        Returns:
            parsed_answer (list of dict): List of Json objects, where the last Json object is the answer.
        """
        parsed_answer, error_msg = extract_json_objects(self.node.result)
        if parsed_answer is None:
            raise AfterQueryError("Failed to parse answer", error_msg)
        elif parsed_answer[-1] is None:
            raise AfterQueryError("No answer", "Invalid Json: It seems that the last Json object in the output above is invalid.")
        elif type(parsed_answer[-1]) != self.type:
            raise AfterQueryError("Invalid answer", "Invalid Type: Expecting the last Json object to be {}, got {} instead.".format(self.type, type(parsed_answer[-1])))
        if self.length is not None and len(parsed_answer[-1]) != self.length:
            raise AfterQueryError("Invalid answer", "Expecting length {}, got {} instead.".format(self.length, len(parsed_answer[-1])))

        for k in self.required_keys:
            if k not in parsed_answer[-1].keys():
                raise AfterQueryError("Invalid answer", "Expecting '{}' in the keys.".format(k))
        
        return parsed_answer


import os
try:
    import anthropic
except ImportError:
    raise ImportError("Please install anthropic to use built-in LLM API.")
from .utils import match_model
import time
from .base import BaseModel

if os.environ.get("ANTHROPIC_KEY") is None:
    print("Environment variable for Anthropic key not found, using Anthropic API key from ~/.openai/openai.key.")
    if not os.path.exists(os.path.join(os.path.expanduser('~'), ".openai/openai.key")):
        raise FileNotFoundError("Please create a file at ~/.openai/openai.key with your OpenAI API key and organization ID. The third line should be your anthropic API key.")
    else:
        with open(os.path.join(os.path.expanduser('~'), ".openai/openai.key"), 'r') as f:
            org_key = f.readlines()
            ANTHROPIC_KEY = org_key[2].strip()
else:
    print("Using Anthropic API key from environment variable.")
    ANTHROPIC_KEY = os.environ.get("ANTHROPIC_KEY")

client = anthropic.Anthropic(
    api_key=ANTHROPIC_KEY,
)

class Claude_chat(BaseModel):

    def __init__(self, model_name, global_counter=None):
        model_max, _ = match_model(model_name)
        self.model_max = model_max
        self.tokenizer = client.get_tokenizer()
        super().__init__(model_name, global_counter, "chat")

    def encode(self, txt):
        return self.tokenizer.encode(txt)

    def decode(self, txt):
        return self.tokenizer.decode(txt)

    def convert_anthropic(self, msg, shrink_idx):
        system = None
        if msg[0]['role'] == 'system':
            system = msg[0]['content']
            msg = msg[1:]
            shrink_idx -= 1
        
        message = []
        role = "user"
        added_terms = 0
        new_index = shrink_idx
        for i,m in enumerate(msg):
            if role == 'user':
                if m['role'] in ('system', 'user'):
                    message.append({"role":"user", "content":m['content']})
                else:
                    message.append({"role":"user", "content":""})
                    added_terms+=1
                    message.append({"role":"assistant", "content":m['content']})
            elif role == 'assistant':
                if m['role'] in ('system', 'user'):
                    message.append({"role":"assistant", "content":"Understood."})
                    added_terms+=1
                    message.append({"role":"user", "content":m['content']})
                else:
                    message.append({"role":"assistant", "content":m['content']})

            if i == shrink_idx:
                new_index = shrink_idx + added_terms
            
            if message[-1]['role'] == 'user':
                role = 'assistant'
            else:
                role = 'user'
        
        return message, system, new_index

    
    def query_chat(self, messages, shrink_idx, max_gen=512, temp=0.):
        model_max = self.model_max
        messages, system, shrink_idx = self.convert_anthropic(messages, shrink_idx)
        messages = self.shrink_msg(messages, shrink_idx, model_max-max_gen)
        while(True):
            try:
                message = client.messages.create(
                    model=self.name,
                    system=system,
                    messages=messages,
                    temperature=temp,
                    max_tokens=max_gen,
                )
                return message.content[0].text, {"prompt":message.usage.input_tokens, "completion":message.usage.output_tokens, "total":message.usage.input_tokens+message.usage.output_tokens}
            except Exception as e:
                if self.debug:
                    raise e
                elif isinstance(e, anthropic.APIConnectionError) or isinstance(e, anthropic.APIStatusError) or isinstance(e, anthropic.InternalServerError):
                    time.sleep(30)
                elif isinstance(e, anthropic.RateLimitError):
                    time.sleep(5*60)
                elif "However, your messages resulted in" in str(e):
                    print("error:", e, str(e))
                    e = str(e)
                    index = e.find("your messages resulted in ")
                    import re
                    val = int(re.findall(r'\d+', e[index + len("your messages resulted in ") : ])[0])
                    index2 = e.find("maximum context length is ")
                    model_max = int(re.findall(r'\d+', e[index2 + len("maximum context length is "):])[0])
                    messages = self.shrink_msg_by(messages, shrink_idx, val-model_max)
                else:
                    time.sleep(5)
                    print(e)